// Gather all the messy details of IL generation together in one place, to
// avoid cluttering up the rest of the codebase.

using System;
using System.Reflection;
using System.Reflection.Emit;



class Compiler
{
    // argument indices for the Function.BodyN prototypes
    public const int RecursionArg = 0;
    public const int GlobalsArg = 1;
    public const int Args = 2;



    // compile a function body
    public static DynamicMethod CompileFunction(ParseTree parseTree, string[] locals)
    {
        Type[] argTypes;

        if (locals.Length <= Function.MaxArgs) {
            // pass fixed arguments directly to a suitably typed delegate
            argTypes = new Type[Args+locals.Length];

            argTypes[RecursionArg] = typeof(int);
            argTypes[GlobalsArg] = typeof(Globals);

            for (int i=0; i<locals.Length; i++)
                argTypes[Args+i] = typeof(Complex);
        }
        else {
            // too many arguments, so pass a dynamically allocated array
            argTypes = new Type[] { typeof(int), typeof(Globals), typeof(Complex[]) };
        }

        // create the method
        DynamicMethod method = new DynamicMethod("elf23", typeof(Complex), argTypes, typeof(Function));

        ILGenerator il = method.GetILGenerator();

        // validate the number of arguments
        if (locals.Length > Function.MaxArgs) {
            Label ok = il.DefineLabel();

            il.Emit(OpCodes.Ldarg, Args);
            il.Emit(OpCodes.Ldlen);
            il.Emit(OpCodes.Ldc_I4, locals.Length);
            il.Emit(OpCodes.Beq, ok);

            il.EmitCall(OpCodes.Call, typeof(Function).GetMethod("ArgCountError", BindingFlags.Static | BindingFlags.NonPublic), null);

            il.MarkLabel(ok);
        }

        // compile the expression parse tree
        Compile(il, parseTree, locals, typeof(Complex));

        il.Emit(OpCodes.Ret);

        return method;
    }



    // look up a function from the symbol table
    internal static Function LookupSymbol(Globals globals, string name)
    {
        Function function;
        
        if (!globals.TryGetValue(name, out function))
            throw new ExpressionException(string.Format("Unknown value {0}.", name));
        
        return function;
    }



    // compile a call to a function
    public static void CallFunction(ILGenerator il, string[] locals, string name, ParseTree[] args)
    {
        // look up the function
        il.Emit(OpCodes.Ldarg, GlobalsArg);
        il.Emit(OpCodes.Ldstr, name);

        il.EmitCall(OpCodes.Call, typeof(Compiler).GetMethod("LookupSymbol", BindingFlags.Static | BindingFlags.NonPublic), null);

        // lookup the appropriate body for this number of arguments
        string fieldname = "mBody" + ((args.Length > Function.MaxArgs) ? "N" : args.Length.ToString());

        FieldInfo body = typeof(Function).GetField(fieldname, BindingFlags.Instance | BindingFlags.NonPublic);

        il.Emit(OpCodes.Ldfld, body);

        // decrement the recursion counter
        il.Emit(OpCodes.Ldarg, RecursionArg);
        il.Emit(OpCodes.Ldc_I4, 1);
        il.Emit(OpCodes.Sub);

        // did we hit zero?
        Label recursionOk = il.DefineLabel();

        il.Emit(OpCodes.Dup);
        il.Emit(OpCodes.Brtrue, recursionOk);

        // throw an exception if we recursed too far
        il.Emit(OpCodes.Ldstr, "Excessive recursion.");
        il.Emit(OpCodes.Newobj, typeof(ExpressionException).GetConstructor(new Type[] { typeof(string) } ));
        il.Emit(OpCodes.Throw);

        il.MarkLabel(recursionOk);

        // load the globals pointer
        il.Emit(OpCodes.Ldarg, GlobalsArg);

        if (args.Length <= Function.MaxArgs) {
            // fast path: evaluate arguments directly onto the stack
            for (int i=0; i<args.Length; i++)
                Compile(il, args[i], locals, typeof(Complex));
        }
        else {
            // slow path: evaluate arguments into a argument array
            il.Emit(OpCodes.Ldc_I4, args.Length);
            il.Emit(OpCodes.Newarr, typeof(Complex));

            for (int i=0; i<args.Length; i++) {
                il.Emit(OpCodes.Dup);
                il.Emit(OpCodes.Ldc_I4, i);
                il.Emit(OpCodes.Ldelema, typeof(Complex));

                Compile(il, args[i], locals, typeof(Complex));

                il.Emit(OpCodes.Stobj, typeof(Complex));
            }
        }

        // call the function
        il.EmitCall(OpCodes.Callvirt, body.FieldType.GetMethod("Invoke"), null);
    }



    // load a function argument onto the stack
    public static void LoadLocal(ILGenerator il, string[] locals, int index)
    {
        if (locals.Length <= Function.MaxArgs) {
            // normal fast path
            il.Emit(OpCodes.Ldarg, Args+index);
        }
        else {
            // functions with too many args use a parameter array
            il.Emit(OpCodes.Ldarg, Args);
            il.Emit(OpCodes.Ldc_I4, index);
            il.Emit(OpCodes.Ldelema, typeof(Complex));
            il.Emit(OpCodes.Ldobj, typeof(Complex));
        }
    }



    // compile an arbitrary parse tree, converting the result to the desired type
    public static void Compile(ILGenerator il, ParseTree parseTree, string[] locals, Type wantType)
    {
        Type resultType = parseTree.Compile(il, locals);

        if (resultType != wantType)
            il.EmitCall(OpCodes.Call, typeof(Ops.Casts).GetMethod(wantType.Name, new Type[] { resultType } ), null);
    }



    // either compile a parse tree or load an integer constant
    static void Compile(ILGenerator il, object obj, string[] locals, Type wantType)
    {
        ParseTree parseTree = obj as ParseTree;

        if (parseTree != null)
            Compile(il, parseTree, locals, wantType);
        else
            il.Emit(OpCodes.Ldc_I4, (int)obj);
    }



    // helper for compiling the lazily evaluated ?, ||, and && operators
    public static Type Ternary(ILGenerator il, string[] locals, ParseTree test, object ifTrue, object ifFalse, Type wantType)
    {
        Label done = il.DefineLabel();
        Label skip = il.DefineLabel();

        // evaluate the test
        Compile(il, test, locals, typeof(bool));

        il.Emit(OpCodes.Brfalse, skip);

        // if true, evaluate the first object
        Compile(il, ifTrue, locals, wantType);

        il.Emit(OpCodes.Br, done);

        // otherwise, evaluate the second object
        il.MarkLabel(skip);

        Compile(il, ifFalse, locals, wantType);

        il.MarkLabel(done);

        return wantType;
    }
}

