// Massive overuse of reflection, mostly just because we can, but also to
// learn how it works.

using System;
using System.Reflection;
using System.Collections.Generic;
using OperatorAttributes;



// metadata tags describe methods of the Ops class
namespace OperatorAttributes
{
	sealed class SymbolAttribute : Attribute
	{
		public SymbolAttribute(string symbol) { Value = symbol; }

		public readonly string Value;
	}


	sealed class PrecedenceAttribute : Attribute
	{
		public PrecedenceAttribute(Ops.Precedence precedence) { Value = precedence; }

		public readonly Ops.Precedence Value;
	}


	sealed class ArityAttribute : Attribute
	{
		public ArityAttribute(int arity) { Value = arity; }

		public readonly int Value;
	}


	sealed class RightAssociativeAttribute : Attribute { }
}



class OpTable
{
	// expose all operators and functions for the tokenizer and parser
	public static readonly Dictionary<string, Operator> Operators = new Dictionary<string, Operator>();
	public static readonly Dictionary<string, Operator> Functions = new Dictionary<string, Operator>();



	// also directly expose a few special operators
	public static readonly Operator OpenBrace    = null;
	public static readonly Operator CloseBrace   = null;
	public static readonly Operator Terminator   = null;
	public static readonly Operator Assign       = null;
	public static readonly Operator Add          = null;
	public static readonly Operator Subtract     = null;
	public static readonly Operator UnaryNegate  = null;
	public static readonly Operator QuestionMark = null;
	public static readonly Operator Colon        = null;
	public static readonly Operator Ternary      = null;



	// let us reflect for a moment...
	static OpTable()
	{
		foreach (MethodInfo method in typeof(Ops).GetMethods(BindingFlags.Public | BindingFlags.Instance | BindingFlags.Static | BindingFlags.DeclaredOnly)) {
			Operator op = ConstructOperator(method);

			// store in the operator or function table
			string symbol = GetSymbol(method);

			if (symbol == null)
				Functions[method.Name] = op;
			else if (symbol.Length > 0)
				Operators[symbol] = op;

			// store in one of the directly exposed special fields?
			if (typeof(OpTable).GetField(method.Name) != null)
				typeof(OpTable).InvokeMember(method.Name, BindingFlags.SetField, null, null, new object [] { op } );
		}
	}



	// wrap a suitable operator class around a handler method
	static Operator ConstructOperator(MethodInfo method)
	{	
		int arity = method.GetParameters().Length;

		Ops.Precedence precedence = GetPrecedence(method, arity);

		bool rightass = GetRightAssociative(method);

		if (method.ReturnType == typeof(void))
			return new SpecialOperator(method.Name, arity, precedence, rightass);

		if (method.ReturnType == typeof(Type))
			return new LazyOperator(method.Name, GetArity(method), precedence, rightass, (LazyOperator.CompileHandler)Delegate.CreateDelegate(typeof(LazyOperator.CompileHandler), method));

		return new StandardOperator(method.Name, arity, precedence, rightass, method);
	}



	// query the operator symbol attribute of a handler method
	static string GetSymbol(MethodInfo method)
	{
		foreach (SymbolAttribute attrib in method.GetCustomAttributes(typeof(SymbolAttribute), false))
			return attrib.Value;

		return null;
	}



	// query the precedence attribute of a handler method
	static Ops.Precedence GetPrecedence(MethodInfo method, int arity)
	{
		foreach (PrecedenceAttribute attrib in method.GetCustomAttributes(typeof(PrecedenceAttribute), false))
			return attrib.Value;

		return (arity == 1) ? Ops.Precedence.Unary : Ops.Precedence.None;
	}



	// query the arity of an lazily evaluated operator
	static int GetArity(MethodInfo method)
	{
		return (method.GetCustomAttributes(typeof(ArityAttribute), false)[0] as ArityAttribute).Value;
	}



	// query the associativity of an operator
	static bool GetRightAssociative(MethodInfo method)
	{
		return (method.GetCustomAttributes(typeof(RightAssociativeAttribute), false).Length > 0);
	}
}

