// OpTable uses these classes to wrap handler methods from the Ops class.

using System;
using System.Reflection;
using System.Reflection.Emit;



// base class stores operator attributes, and knows how to compile itself
abstract class Operator
{
	protected Operator(string name, int arity, Ops.Precedence precedence, bool rightass)
	{
		Name = name;
		Arity = arity;
		Precedence = precedence;
		RightAssociative = rightass;
	}


	public readonly string Name;
	public readonly int Arity;
	public readonly Ops.Precedence Precedence;
	public readonly bool RightAssociative;


	abstract public Type Compile(ILGenerator il, string[] locals, ParseTree[] args);
}



// special operators like the terminator should never actually be evaluated
class SpecialOperator : Operator
{
	public SpecialOperator(string name, int arity, Ops.Precedence precedence, bool rightass)
		: base(name, arity, precedence, rightass)
		{ }


	override public Type Compile(ILGenerator il, string[] locals, ParseTree[] args)
	{
		throw new ExpressionException(string.Format("Invalid use of {0} operator.", Name));
	}
}



// a regular nullary, unary, or binary operator
class StandardOperator : Operator
{
	public StandardOperator(string name, int arity, Ops.Precedence precedence, bool rightass, MethodInfo handler)
		: base(name, arity, precedence, rightass)
	{
		mHandler = handler;
	}


	MethodInfo mHandler;


	override public Type Compile(ILGenerator il, string[] locals, ParseTree[] args)
	{
		if (args.Length != Arity)
			throw new ExpressionException(string.Format("Wrong number of arguments for {0} function.", Name));

		for (int i=0; i<args.Length; i++)
			Compiler.Compile(il, args[i], locals, mHandler.GetParameters()[i].ParameterType);

		il.EmitCall(OpCodes.Call, mHandler, null);

		return mHandler.ReturnType;
	}
}



// lazily evaluated operators provide their own special compile functions
class LazyOperator : Operator
{
	public LazyOperator(string name, int arity, Ops.Precedence precedence, bool rightass, CompileHandler compiler)
		: base(name, arity, precedence, rightass)
	{
		mCompiler = compiler;
	}


	public delegate Type CompileHandler(ILGenerator il, string[] locals, ParseTree[] args);

	CompileHandler mCompiler;


	override public Type Compile(ILGenerator il, string[] locals, ParseTree[] args)
	{
		if (args.Length != Arity)
			throw new ExpressionException(string.Format("Wrong number of arguments for lazy {0} operator.", Name));

		return mCompiler(il, locals, args);
	}
}

