// Builtin operator implementations. OpTable reflects over this class and
// instantiates an Operator object for each public method.

using System;
using System.Reflection.Emit;
using OperatorAttributes;



abstract class Ops
{
	// precedence table
	public enum Precedence
	{
		None,
		Brace,
		Terminator,
		Assign,
		Ternary,
		LogicalOr,
		LogicalXor,
		LogicalAnd,
		BinaryOr,
		BinaryXor,
		BinaryAnd,
		CompareEq,
		CompareDiff,
		Shift,
		Addition,
		Multiply,
		Unary,
		Power,
	};



	// special marker operators don't need implementations
	[Symbol("("), Precedence(Precedence.Brace)]
	public abstract void OpenBrace();

	[Symbol(")"), Precedence(Precedence.Brace)]
	public abstract void CloseBrace();

	[Symbol(""), Precedence(Precedence.Terminator)]
	public abstract void Terminator();

	[Symbol("="), Precedence(Precedence.Assign)]
	public abstract void Assign(Complex x, Complex y);



	// component parts of the ternary ?: operator
	[Symbol("?"), Precedence(Precedence.Ternary), RightAssociative()]
	public abstract void QuestionMark(Complex x, Complex y);

	[Symbol(":"), Precedence(Precedence.Ternary), RightAssociative()]
	public abstract void Colon(Complex x, Complex y);



	// lazily evaluated logical operators
	[Symbol(""), Precedence(Precedence.Ternary), Arity(3)]
	public static Type Ternary(ILGenerator il, string[] locals, ParseTree[] args)
	{
		return Compiler.Ternary(il, locals, args[0], args[1], args[2], typeof(Complex));
	}

	[Symbol("||"), Precedence(Precedence.LogicalOr), Arity(2)]
	public static Type LogicalOr(ILGenerator il, string[] locals, ParseTree[] args)
	{
		return Compiler.Ternary(il, locals, args[0], 1, args[1], typeof(bool));
	}

	[Symbol("&&"), Precedence(Precedence.LogicalAnd), Arity(2)]
	public static Type LogicalAnd(ILGenerator il, string[] locals, ParseTree[] args)
	{
		return Compiler.Ternary(il, locals, args[0], args[1], 0, typeof(bool));
	}



	// binary operators
	[Symbol("|"), Precedence(Precedence.BinaryOr)]
	public static int BinaryOr(int x, int y)
	{
		return x | y;
	}

	[Symbol("^^"), Precedence(Precedence.BinaryXor)]
	public static int BinaryXor(int x, int y)
	{
		return x ^ y;
	}

	[Symbol("&"), Precedence(Precedence.BinaryAnd)]
	public static int BinaryAnd(int x, int y)
	{
		return x & y;
	}



	// comparisons
	[Symbol("=="), Precedence(Precedence.CompareEq)]
	public static bool Equal(double x, double y)
	{
		return x == y;
	}

	[Symbol("!="), Precedence(Precedence.CompareEq)]
	public static bool NotEqual(double x, double y)
	{
		return x != y;
	}

	[Symbol("<"), Precedence(Precedence.CompareDiff)]
	public static bool Less(double x, double y)
	{
		return x < y;
	}

	[Symbol(">"), Precedence(Precedence.CompareDiff)]
	public static bool Greater(double x, double y)
	{
		return x > y;
	}

	[Symbol("<="), Precedence(Precedence.CompareDiff)]
	public static bool LessOrEqual(double x, double y)
	{
		return x <= y;
	}

	[Symbol(">="), Precedence(Precedence.CompareDiff)]
	public static bool GreaterOrEqual(double x, double y)
	{
		return x >= y;
	}



	// shifts
	[Symbol("<<"), Precedence(Precedence.Shift)]
	public static int LeftShift(int x, int y)
	{
		return x << y;
	}

	[Symbol(">>"), Precedence(Precedence.Shift)]
	public static double RightShift(int x, int y)
	{
		return (uint)x >> y;
	}

	[Symbol(">>>"), Precedence(Precedence.Shift)]
	public static int ArithmeticRightShift(int x, int y)
	{
		return x >> y;
	}



	// arithmetic
	[Symbol("+"), Precedence(Precedence.Addition)]
	public static Complex Add(Complex x, Complex y)
	{
		return x + y;
	}

	[Symbol("-"), Precedence(Precedence.Addition)]
	public static Complex Subtract(Complex x, Complex y)
	{
		return x - y;
	}

	[Symbol("*"), Precedence(Precedence.Multiply)]
	public static Complex Multiply(Complex x, Complex y)
	{
		return x * y;
	}

	[Symbol("/"), Precedence(Precedence.Multiply)]
	public static Complex Divide(Complex x, Complex y)
	{
		return x / y;
	}

	[Symbol("%"), Precedence(Precedence.Multiply)]
	public static double Modulus(double x, double y)
	{
		double ret = Math.IEEERemainder(x, y);

		if (ret < 0)
			ret += y;

		return ret;
	}



	// negation
	[Symbol("!")]
	public static bool LogicalNot(bool x)
	{
		return !x;
	}

	[Symbol("~")]
	public static int BinaryNot(int x)
	{
		return ~x;
	}

	[Symbol("")]
	public static Complex UnaryNegate(Complex x)
	{
		return -x;
	}



	// raise to a power
	[Symbol("^"), Precedence(Precedence.Power)]
	public static double Power(double x, double y)
	{
		return Math.Pow(x, y);
	}



	// math functions
	public static double max(double x, double y)	{ return Math.Max(x, y);  }
	public static double min(double x, double y)	{ return Math.Min(x, y);  }
	public static double sqrt(double x)				{ return Math.Sqrt(x);    }
	public static double exp(double x)				{ return Math.Exp(x);     }
	public static double ln(double x)				{ return Math.Log(x);     }
	public static double log(double x)				{ return Math.Log10(x);   }
	public static double ceil(double x)				{ return Math.Ceiling(x); }
	public static double floor(double x)			{ return Math.Floor(x);   }



	// trig
	public static double sin(double x)		{ return Math.Sin(Radians(x));  }
	public static double cos(double x)		{ return Math.Cos(Radians(x));  }
	public static double tan(double x)		{ return Math.Tan(Radians(x));  }
	public static double sinh(double x)		{ return Math.Sinh(Radians(x)); }
	public static double cosh(double x)		{ return Math.Cosh(Radians(x)); }
	public static double tanh(double x)		{ return Math.Tanh(Radians(x)); }
	public static double asin(double x)		{ return Angle(Math.Asin(x));   }
	public static double acos(double x)		{ return Angle(Math.Acos(x));   }
	public static double atan(double x)		{ return Angle(Math.Atan(x));   }



	// integer casts
	public static int    s8(int x)			{ return (sbyte) x; }
	public static int    u8(int x)			{ return (byte)  x; }
	public static int    s16(int x)			{ return (short) x; }
	public static int    u16(int x)			{ return (ushort)x; }
	public static int    s32(int x)			{ return (int)   x; }
	public static double u32(int x)			{ return (uint)  x; }



	// constants
	public static double e()				{ return Math.E;  }
	public static double pi()				{ return Math.PI; }
	public static double circle()			{ return Circle;  }



	// complex numbers
	public static double real(Complex x)	{ return x.Real; }
	public static double imag(Complex x)	{ return x.Imag; }
	public static Complex abs(Complex x)	{ return x.Abs(); }
	public static Complex i()				{ return new Complex(0, 1); }



	// type conversions
	public class Casts
	{
		public static Complex Complex(double x)		{ return x; }
		public static Complex Complex(int x)		{ return x; }
		public static Complex Complex(bool x)		{ return x ? 1 : 0; }

		public static double Double(int x)			{ return x; }
		public static double Double(bool x)			{ return x ? 1 : 0; }

		public static int Int32(Complex x)			{ return Int32(Double(x)); }
		public static int Int32(double x)			{ return (int)Math.Round(x); }
		public static int Int32(bool x)				{ return x ? 1 : 0; }

		public static bool Boolean(Complex x)		{ return Double(x) != 0; }
		public static bool Boolean(double x)		{ return x != 0; }
		public static bool Boolean(int x)			{ return x != 0; }

		public static double Double(Complex x)
		{
			if (x.Imag != 0)
				throw new InvalidCastException("Operation does not support complex numbers.");

			return x.Real;
		}
	}



	// current units of angle, and conversions to/from radians
	public const double Pi2 = Math.PI*2;

	public static double Circle = Pi2;

	static double Radians(double x)	{ return x * Pi2 / Circle; }
	static double Angle(double x)	{ return x * Circle / Pi2; }
}

