// The parser creates this tree of expression nodes, which can then be
// compiled into an executable function.

using System;
using System.Reflection.Emit;



abstract class ParseTree
{
	// generate IL code
	abstract public Type Compile(ILGenerator il, string[] locals);



	// attach a source tag for debugging or UI display
	public string SourceCode;



	// helper for formatting argument lists
	static protected string ToString(ParseTree[] args)
	{	
		if (args.Length == 0)
			return "";

		return '(' + string.Join(", ",
						Array.ConvertAll<ParseTree, string>(args,
							delegate(ParseTree e) { return e.ToString(); } )) + ')';
	}
}



// a constant numeric value
class ParseTreeConstant : ParseTree
{
	public ParseTreeConstant(double value)
	{
		mValue = value;
	}



	// the constant value
	double mValue;



	// generate IL code
	override public Type Compile(ILGenerator il, string[] locals)
	{
		il.Emit(OpCodes.Ldc_R8, mValue);

		return typeof(double);
	}



	// format ourselves for debugging
	override public string ToString()
	{
		return mValue.ToString();
	}
}



// call a builtin operator
class ParseTreeOperator : ParseTree
{
	public ParseTreeOperator(Operator op, ParseTree[] args)
	{
		Operator = op;
		Args = args;
	}



	// operator plus arguments
	public readonly Operator Operator;

	public readonly ParseTree[] Args;



	// generate IL code
	override public Type Compile(ILGenerator il, string[] locals)
	{
		return Operator.Compile(il, locals, Args);
	}



	// format ourselves for debugging
	override public string ToString()
	{
		return Operator.Name + ToString(Args);
	}
}



// call a user defined function
class ParseTreeFunction : ParseTree
{
	public ParseTreeFunction(string name, ParseTree[] args)
	{
		Name = name;
		Args = args;
	}



	// function name and arguments
	public readonly string Name;

	public readonly ParseTree[] Args;



	// generate IL code
	override public Type Compile(ILGenerator il, string[] locals)
	{
		int index = Array.IndexOf(locals, Name);

		if (index >= 0) {
			// look up a local function argument
			if (Args.Length > 0)
				throw new ExpressionException("First class functions are not supported.");

			Compiler.LoadLocal(il, locals, index);
		}
		else {
			// call a different function
			Compiler.CallFunction(il, locals, Name, Args);
		}

		return typeof(Complex);
	}



	// format ourselves for debugging
	override public string ToString()
	{
		return Name + ToString(Args);
	}
}

