// The main interpreter object. Given a line of code, uses reflection to
// find an appropriate handler method, or passes it to the function parser.

using System;
using System.IO;
using System.Collections.Generic;
using System.Reflection;



class Interpreter
{
	// user defined functions, with undo data
	Globals mGlobals = new Globals();

	Stack<Undo> mUndo = new Stack<Undo>();
	Stack<Undo> mRedo = new Stack<Undo>();



	// entrypoint for evaluating a line of code
	public bool Evaluate(string code)
	{
		TokenPeeker tokenizer = new TokenPeeker(new Tokenizer(code, OpTable.Operators));

		// dispatch command via reflection?
		if (tokenizer.Peek() is string) {
			bool result;

			if (ReflectiveDispatch(tokenizer, (string)tokenizer.Peek(), out result))
				return result;
		}

		// parse expressions
		while (tokenizer.Peek() != null) {
			ParseTree parseTree = new Parser(tokenizer, false).Result;

			if (mDebug)
				Console.WriteLine(parseTree);
			else
				DefineFunction(parseTree);
		}

		return true;
	}



	// compile an expression into a function
	void DefineFunction(ParseTree parseTree)
	{
		string name;
		string[] args;
		ParseTree body;

		if (Function.Munge(parseTree, out name, out args, out body)) {
			// define a new function
			mUndo.Push(new Undo(mGlobals, name));
			mRedo.Clear();

			mGlobals[name] = new Function(body, args);
		}
		else {
			// immediate evaluation mode
			Complex result = new Function(parseTree, null).Evaluate(mGlobals);

			Console.WriteLine(Format.Pretty(result));
		}
	}



	// dispatch a command to a suitable handler method, if any exists
	bool ReflectiveDispatch(TokenPeeker tokenizer, string command, out bool result)
	{
		result = true;

		// reflect to find a handler
		MethodInfo handler = GetType().GetMethod(command+"Command", BindingFlags.Instance | BindingFlags.Static | BindingFlags.NonPublic);

		if (handler == null)
			return false;

		tokenizer.Read();

		// does the command want any further arguments?
		object[] args;

		if (handler.GetParameters().Length == 0) {
			if (tokenizer.Peek() != null)
				throw new UsageException("Usage: " + command);

			args = null;
		}
		else
			args = new object[] { tokenizer };

		// dynamic method dispatch
		object ret = handler.Invoke(this, args);

		if (ret is bool)
			result = (bool)ret;

		return true;
	}



	// query whether a command is currently valid, using reflection
	public bool CommandQuery(string command)
	{
		MethodInfo handler = GetType().GetMethod(command+"Valid", BindingFlags.Instance | BindingFlags.Static | BindingFlags.NonPublic);

		if (handler == null)
			return true;

		return (bool)handler.Invoke(this, null);
	}



	// exit handlers
	static bool exitCommand()	{ return false; }
	static bool quitCommand()	{ return false; }
	static bool byeCommand ()	{ return false; }
	static bool qCommand   ()	{ return false; }



	// undo/redo handlers
	void undoCommand() { Undo.Perform(mUndo, mRedo, mGlobals); }
	void redoCommand() { Undo.Perform(mRedo, mUndo, mGlobals); }



	// query command validity
	bool undoValid() { return (mUndo.Count > 0); }
	bool redoValid() { return (mRedo.Count > 0); }



	// read code from a text file
	void loadCommand(TokenPeeker tokenizer)
	{
		string filename = tokenizer.Read() as string;

		if ((filename == null) || (tokenizer.Peek() != null))
			throw new UsageException("Usage: load <filename>");

		Console.WriteLine("Loading {0}.", filename);

		using (StreamReader input = new StreamReader(filename)) {
			foreach (string line in new LineEnumerator(input)) {
				if (!Evaluate(line))
					break;
			}
		}
	}



	// write definitions to a text file
	void saveCommand(TokenPeeker tokenizer)
	{
		string filename = tokenizer.Read() as string;

		if ((filename == null) || (tokenizer.Peek() != null))
			throw new UsageException("Usage: save <filename>");

		using (StreamWriter output = new StreamWriter(filename)) {
			foreach (string var in Util.Sort(mGlobals.Keys))
				output.WriteLine(mGlobals[var].SourceCode);
		}

		Console.WriteLine("Saved {0}.", filename);
	}



	// list the current definitions
	void lsCommand()
	{
		foreach (string var in Util.Sort(mGlobals.Keys))
			Console.WriteLine(mGlobals[var].SourceCode);
	}



	// set angle units to degrees
	static void degCommand()
	{
		Console.WriteLine("Angles in degrees.");

		Ops.Circle = 360;
	}



	// set angle units to radians
	static void radCommand()
	{
		Console.WriteLine("Angles in radians.");

		Ops.Circle = Ops.Pi2;
	}



	// change or display the number base
	static void baseCommand(TokenPeeker tokenizer)
	{
		List<int> bases = new List<int>();

		foreach (object t in tokenizer) {
			if (t is double)
				bases.Add((int)(double)t);
			else
				throw new UsageException("Usage: base <list of number bases>");
		}

		if (bases.Count > 0)
			Format.Bases = bases.ToArray();

		// display the base settings
		List<string> s = new List<string>();

		foreach (int b in Format.Bases)
			s.Add(b.ToString());

		Console.WriteLine("Using base {0}.", string.Join(" ", s.ToArray()));
	}



	// display the help page (written in a functional style just for kicks)
	void helpCommand()
	{
		Util.WriteHelp("Operators",
			Array.ConvertAll<string, string>(
				Util.Sort(OpTable.Operators.Keys),
				delegate(string op) {
					return op.PadRight(5) + OpTable.Operators[op].Name;
				}
			)
		);

		Util.WriteHelp("Functions", Util.Sort(OpTable.Functions.Keys));

		Util.WriteHelp("Commands", 
			Util.Sort(
				Array.ConvertAll<MethodInfo, string>(
					Array.FindAll(
						GetType().GetMethods(BindingFlags.Instance | BindingFlags.Static | BindingFlags.NonPublic),
						delegate(MethodInfo cmd) {
							return cmd.Name.EndsWith("Command");
						}
					),
					delegate(MethodInfo cmd) {
						return cmd.Name.Substring(0, cmd.Name.Length - "Command".Length);
					}
				)
			)
		);
	}



	// turn parser debug mode on/off
	bool mDebug;

	void debugCommand(TokenPeeker tokenizer)
	{
		object debug = tokenizer.Read();

		if (debug == null)
			mDebug = true;
		else if (!(debug is double) || (tokenizer.Peek() != null))
			throw new UsageException("Usage: debug <enable>");
		else
			mDebug = (double)debug != 0;

		Console.WriteLine("Parse tree debugging = {0}.", mDebug ? 1 : 0);
	}



	// add expressions to the watch window
	void watchCommand(TokenPeeker tokenizer)
	{
		bool append;

		List<ParseTree> parseTrees = ParseExpressions(tokenizer, out append, "watch", "");

		List<Function> functions = new List<Function>();

		foreach (ParseTree parseTree in parseTrees)
			functions.Add(new Function(parseTree, null));

		Watch.Add(functions, mGlobals, append);
	}



	// graph plotter
	void plotCommand(TokenPeeker tokenizer)
	{
		bool append;

		List<ParseTree> parseTrees = ParseExpressions(tokenizer, out append, "plot", "\n\n" + Plot.Usage);

		Plot.Add(parseTrees, mGlobals, append);
	}



	// helper for parsing arguments of the watch or plot commands
	static List<ParseTree> ParseExpressions(TokenPeeker tokenizer, out bool append, string command, string usageTail)
	{
		append = false;

		if (tokenizer.Peek() == OpTable.Add) {
			tokenizer.Read();
			append = true;
		}

		if (tokenizer.Peek() == null)
			throw new UsageException("Usage: " + command + " [+] <expression> { , <expression> }" + usageTail);

		List<ParseTree> parseTrees = new List<ParseTree>();

		while (tokenizer.Peek() != null)
			parseTrees.Add(new Parser(tokenizer, false).Result);

		return parseTrees;
	}
}

