// The program entrypoint and read/eval loop.

using System;
using System.IO;
using System.Reflection;

[assembly: AssemblyVersion("1.0.*")]



class EricApp
{
	// entrypoint
	[STAThreadAttribute()]
	static int Main(string[] args)
	{
		int exitcode = 0;

		Interpreter interpreter = new Interpreter();

		TextReader input = SelectInput(args, interpreter);

		// read/eval loop
		foreach (string line in new LineEnumerator(input)) {
			if (!Evaluate(line, interpreter, input, ref exitcode))
				break;
		}

		EricForm.Shutdown(input is LineEditor);

		return exitcode;
	}



	// choose where to get our input
	static TextReader SelectInput(string[] args, Interpreter interpreter)
	{
		if ((args.Length == 1) && (File.Exists(args[0]))) {
			// read from a named file
			return new StreamReader(args[0]);
		}
		else if (args.Length > 0) {
			// read from commandline arguments
			return new StringReader(string.Join(" ", args).Replace(';', '\n'));
		}
		else if (Conio.InputRedirected()) {
			// read from stdin
			return Console.In;
		}
		else {
			// read from an interactive console
			return MakeLineEditor(interpreter);
		}
	}



	// wire up an interactive console reader
	static LineEditor MakeLineEditor(Interpreter interpreter)
	{
		LineEditor editor = new LineEditor();

		editor.AddShortcut(ConsoleKey.F1,     0,                        "help");
		editor.AddShortcut(ConsoleKey.Escape, 0,                        "exit");
		editor.AddShortcut(ConsoleKey.D,      ConsoleModifiers.Control, "exit");
		editor.AddShortcut(ConsoleKey.Z,      ConsoleModifiers.Control, "undo");
		editor.AddShortcut(ConsoleKey.Y,      ConsoleModifiers.Control, "redo");
		editor.AddShortcut(ConsoleKey.Tab,    ConsoleModifiers.Control, EricForm.CycleFocus);

		editor.ShortcutQuery += interpreter.CommandQuery;

		return editor;
	}



	// wrap suitable error handling around Interpreter.Evaluate
	static bool Evaluate(string line, Interpreter interpreter, TextReader input, ref int exitcode)
	{
		#if DEBUG

			// in debug builds, errors go straight to the debugger
			if (!interpreter.Evaluate(line))
				return false;

		#else

			// otherwise trap everything
			try {
				if (!interpreter.Evaluate(line))
					return false;
			}
			catch (Exception error) {
				while (error.InnerException != null)
					error = error.InnerException;

				Console.Error.WriteLine(error.Message);

				// if not in interactive mode, halt on error
				if (!(input is LineEditor)) {
					exitcode = 1;
					return false;
				}
			}

		#endif

		return true;
	}
}

