// Window for plotting curves. Or even straight lines, but those are boring.

using System;
using System.Windows.Forms;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Collections.Generic;
using System.Collections;



class Plot : EricForm
{
    // singleton accessor
    public static void Add(List<ParseTree> parseTrees, Globals globals, bool append)
    {
        Pair<Pair<Function>> ranges = new Pair<Pair<Function>>();

        List<Curve> curves = ParseArgs(parseTrees, ranges);

        Plot plot = Instance<Plot>();

        if (plot == null)
            plot = new Plot();

        plot.Add(curves, ranges, append);
        plot.Show();
        plot.Recalculate(globals);
    }



    // private constructor
    Plot() : base("Plot", 320, 32, 640, 640)
    {
        SizeGripStyle = SizeGripStyle.Show;

        DoubleBuffered = true;

        // recalculate on resize
        Layout += delegate {
            if (mWatching != null)
                Recalculate(mWatching);
        };

        // display the mouse position
        MouseMove += delegate(object sender, MouseEventArgs e) {
            mMousePos = e.Location;
            Invalidate();
        };

        MouseLeave += delegate {
            mMousePos = null;
            Invalidate();
        };
    }



    // a curve, plus the data from evaluating it
    class CurveData
    {
        public CurveData(Curve curve, Color color)
        {
            Curve = curve;
            Color = color;
        }

        public readonly Curve Curve;
        public readonly Color Color;

        public object Data;

        public object Recalc;
        public float RecalcPos;

        public string Error;
    }



    // list of curves
    List<CurveData> mCurves = new List<CurveData>();



    // axis ranges, both requested and current
    Pair<Pair<Function>> mWantRange = new Pair<Pair<Function>>();

    Pair<Range> mShowRange;

    string mRangeError;



    // are we busy thinking?
    bool mBusy;



    // current mouse position
    Point? mMousePos;



    // possible curve colors
    static Color[] mCurveColors =
    {
        Color.Blue,
        Color.Red,
        Color.Green,
        Color.Orange,
        Color.Magenta,
        Color.DarkCyan,
        Color.Black,
        Color.White,
    };



    // add new curves to the display
    void Add(List<Curve> curves, Pair<Pair<Function>> ranges, bool append)
    {
        if (curves.Count == 0)
            append = true;

        if (!append)
            mCurves.Clear();

        foreach (Curve curve in curves)
            mCurves.Add(new CurveData(curve, mCurveColors[mCurves.Count % mCurveColors.Length]));

        for (int i=0; i<2; i++) {
            if ((ranges[i] != null) || (!append))
                mWantRange[i] = ranges[i];
        }
    }



    // recompute the contents of all curves
    protected override void Recalculate(Globals globals)
    {
        base.Recalculate(globals);

        // make copies for safe use by the background thread
        Globals globalsCopy = new Globals(globals);

        int w = Math.Max(ClientSize.Width,  1);
        int h = Math.Max(ClientSize.Height, 1);

        // input and output axis ranges
        Pair<Pair<Function>> wantRange = new Pair<Pair<Function>>(mWantRange);

        Pair<Range> range = new Pair<Range>();

        string rangeError = null;

        // sort the curve list so the ones with the most axis influence go first
        List<CurveData> curves = new List<CurveData>(mCurves);

        curves.Sort(delegate(CurveData a, CurveData b) { return b.Curve.AutoRange - a.Curve.AutoRange; } );

        Dictionary<Curve, object> results = new Dictionary<Curve, object>();
        Dictionary<Curve, string> errors  = new Dictionary<Curve, string>();

        // threaded closures kick ass!
        mWorkerThread.BackgroundTask(this,
            delegate {
                // evaluate any explicit range limits
                for (int i=0; i<2; i++) {
                    if (wantRange[i] != null) {
                        try {
                            range[i] = new Range(wantRange[i].X.Evaluate(globalsCopy).Real,
                                                 wantRange[i].Y.Evaluate(globalsCopy).Real);

                            range[i].Lock();
                        }
                        catch (Exception error) {
                            rangeError = error.Message;
                        }
                    }
                }

                // evaluate the curve functions
                foreach (CurveData curve in curves) {
                    try {
                        results[curve.Curve] = curve.Curve.Evaluate(globalsCopy, w, h, range, curve.Color, CurveProgress);
                    }
                    catch (Exception error) {
                        errors[curve.Curve] = error.Message;
                    }
                }
            },

            // tell the user if this is taking a long time
            delegate {
                mBusy = true;
                Invalidate();
            },

            // readback results in the context of the main thread
            delegate {
                foreach (CurveData curve in mCurves) {
                    SafeDispose(curve.Data);
                    SafeDispose(curve.Recalc);

                    results.TryGetValue(curve.Curve, out curve.Data);
                    errors .TryGetValue(curve.Curve, out curve.Error);

                    curve.Recalc = null;
                    curve.RecalcPos = 0;
                }

                mShowRange = range;
                mRangeError = rangeError;

                mBusy = false;

                Invalidate();
            }
        );
    }



    // worker thread callback displays partial results
    void CurveProgress(Curve curve, Object data, float pos)
    {
        BeginInvoke(new MethodInvoker(
            delegate {
                foreach (CurveData info in mCurves) {
                    if (info.Curve == curve) {
                        SafeDispose(info.Recalc);

                        info.Recalc = data;
                        info.RecalcPos = pos;

                        Invalidate();
                        break;
                    }
                }
            }
        ));
    }



    // helper for getting rid of temporary objects
    static void SafeDispose(object obj)
    {
        IDisposable d = obj as IDisposable;

        if (d != null)
            d.Dispose();
    }



    // clear the window background
    protected override void OnPaintBackground(PaintEventArgs e)
    {
        e.Graphics.Clear(Color.LightGray);
    }



    // render to the screen
    protected override void OnPaint(PaintEventArgs e)
    {
        Graphics gfx = e.Graphics;

        Range xrange, yrange;

        Matrix transform = RangeTransform(out xrange, out yrange);

        if (mShowRange != null)
            PaintAxes(gfx, xrange, yrange, transform);

        gfx.SmoothingMode = SmoothingMode.AntiAlias;

        foreach (CurveData curve in mCurves)
            curve.Curve.Paint(gfx, ClientSize, transform, curve.Color, curve.Data, curve.Recalc, curve.RecalcPos);

        PaintLabels(gfx, transform);

        base.OnPaint(e);
    }



    // set up the range transform
    Matrix RangeTransform(out Range xrange, out Range yrange)
    {
        if (mShowRange != null) {
            xrange = (mShowRange.X == null) ? new Range() : mShowRange.X;
            yrange = (mShowRange.Y == null) ? new Range() : mShowRange.Y;
        }
        else {
            xrange = new Range();
            yrange = new Range();
        }

        Matrix transform = new Matrix();

        transform.Scale(ClientSize.Width  / (float)xrange.Size,
                       -ClientSize.Height / (float)yrange.Size);

        transform.Translate(-(float)xrange.Min,
                            -(float)yrange.Max);

        return transform;
    }



    // render the 2d axis grid
    void PaintAxes(Graphics gfx, Range xrange, Range yrange, Matrix transform)
    {
        int w = ClientSize.Width;
        int h = ClientSize.Height;

        Point[] origin = { new Point(0, 0) };

        transform.TransformPoints(origin);

        // x axis
        PaintAxis(gfx, xrange, w, h, origin[0].Y, StringFormatFlags.DirectionVertical,
            delegate(int x, int y) { return new Point(x, y); }
        );

        // y axis
        PaintAxis(gfx, yrange, h, w, origin[0].X, 0,
            delegate(int x, int y) { return new Point(y, h-x-1); }
        );
    }



    // axis colors
    static Brush mAxisBrush = new SolidBrush(Color.FromArgb(255, 128, 128, 128));
    static Pen   mAxisPen   = new Pen       (Color.FromArgb(192, 255, 255, 255));
    static Pen   mAxisPen2  = new Pen       (Color.FromArgb(64,  255, 255, 255));



    // flip x and y when painting the y axis
    delegate Point AxisHandler(int x, int y);



    // render lines and labels along either the x or y axis
    static void PaintAxis(Graphics gfx, Range range, int w, int h, int y, StringFormatFlags align, AxisHandler chooseAxis)
    {
        // choose label alignment
        StringFormat stringFormat = new StringFormat(align);

        stringFormat.Alignment = StringAlignment.Near;
        stringFormat.LineAlignment = StringAlignment.Center;

        if (y < 0) {
            // clamp to the start of the other axis
            y = 0;
        }
        else if (y >= h) {
            // clamp to the end of the other axis
            stringFormat.Alignment = StringAlignment.Far;
            y = h - 5;
        }
        else {
            // render this axis line
            gfx.DrawLine(mAxisPen, chooseAxis(0, y), chooseAxis(w, y));
        }

        // normalize negative ranges
        double min = Math.Min(range.Min, range.Max);
        double max = Math.Max(range.Min, range.Max);

        double size = Math.Abs(range.Size);

        if ((size < 1e-12) || (size > 1e12))
            return;

        // how far apart should we put the ticks?
        double gap = Math.Pow(10, Math.Round(Math.Log10(size)));

        int fraction = 0;

        while (size/gap < w/48) {
            // 10, 5, 2, 1, 0.5, 0.2, 0.1, etc
            switch (fraction) {
                case 0: gap /= 2;   fraction = 1; break;
                case 1: gap /= 2.5; fraction = 2; break;
                case 2: gap /= 2;   fraction = 0; break;
            }
        }

        // how many fractional digits should we display?
        double digits = Math.Floor(Math.Log10(gap));

        string labelFormat;

        if (Math.Abs(digits) > 5)
            labelFormat = (digits < 0) ? "g" : "e1";
        else
            labelFormat = "f" + Math.Max(0, -digits);

        // round down to the previous tick position
        double start = Math.Floor(min/gap) * gap;

        for (double pos = start; pos <= max; pos += gap) {
            // don't bother labelling the origin
            if (Math.Abs(pos) < gap/2)
                continue;

            // display the tick
            int x = (int)(w * (pos-min) / size);

            gfx.DrawLine(mAxisPen2, chooseAxis(x, 0), chooseAxis(x, h));

            gfx.DrawString(pos.ToString(labelFormat), SystemFonts.DialogFont, mAxisBrush, chooseAxis(x, y+4), stringFormat);
        }
    }



    // draw the expression key, plus any evaluation error messages
    void PaintLabels(Graphics gfx, Matrix transform)
    {   
        int line = 0;

        // curve expressions
        foreach (CurveData curve in mCurves) {
            string title = curve.Curve.SourceCode;

            if (curve.Error != null)
                title += " : " + curve.Error;

            using (Brush brush = new SolidBrush(curve.Color))
                gfx.DrawString(title, SystemFonts.DialogFont, brush, 0, SystemFonts.DialogFont.Height * line++);
        }

        // range evaluation error?
        if (mRangeError != null)
            gfx.DrawString("Range: " + mRangeError, SystemFonts.DialogFont, Brushes.Black, 0, SystemFonts.DialogFont.Height * line++);

        // mouse position?
        if ((mMousePos != null) && (mShowRange != null) && (transform.IsInvertible)) {
            PointF[] p = { new PointF(mMousePos.Value.X, mMousePos.Value.Y) };

            transform.Invert();
            transform.TransformPoints(p);

            gfx.DrawString("x: " + Format(p[0].X), SystemFonts.DialogFont, Brushes.Black, 0, SystemFonts.DialogFont.Height * ++line);
            gfx.DrawString("y: " + Format(p[0].Y), SystemFonts.DialogFont, Brushes.Black, 0, SystemFonts.DialogFont.Height * ++line);
        }

        // busy indicatr?
        if (mBusy) {
            StringFormat alignment = new StringFormat();

            alignment.Alignment = StringAlignment.Center;
            alignment.LineAlignment = StringAlignment.Center;

            gfx.DrawString("Thinking...", SystemFonts.CaptionFont, Brushes.Black, ClientSize.Width/2, ClientSize.Height/2, alignment);
        }
    }



    // guess how many fractional digits to display
    public static string Format(double val)
    {
        double digits = Math.Floor(Math.Log10(Math.Abs(val)));

        if (Math.Abs(digits) < 12)
            return val.ToString("f" + Math.Max(0, -digits+2));
        else
            return val.ToString();
    }



    // usage help text
    public const string Usage =

        "Plot expressions:\n"                     +
        "\n"                                      +
        "    Basic:      f(x)\n"                  +
        "    X axis:     y=f(x)\n"                +
        "    Y axis:     x=f(y)\n"                +
        "    Parametric: x=f(t), y=f(t), t=0:1\n" +
        "    Cartesian:  z=f(x,y)\n"              +
        "    Sets:       s=f(x,y)\n"              +
        "    Ranges:     0:1, x=0:1, y=0:1";



    // helper for turning input expressions into curves and ranges
    static List<Curve> ParseArgs(List<ParseTree> parseTrees, Pair<Pair<Function>> ranges)
    {
        Stack curves = new Stack();

        foreach (ParseTree parseTree in parseTrees) {
            string name;
            string[] locals;
            ParseTree body;

            // figure out whether we have f(x), z=f(x, y), etc
            if (!Function.Munge(parseTree, out name, out locals, out body))
                body = parseTree;

            if ((locals != null) && (locals.Length > 0))
                throw new UsageException(Usage);

            // create an appropriate range or curve object
            if (!ParseRange(body, name, ranges, curves)) {
                switch (name) {

                    case null:
                    case "y":
                        // can't actually construct a YCurve here, because
                        // this could be just the first half of a parametric
                        // curve, so we store parameters for later use
                        curves.Push(new object[] { typeof(YCurve), body } );
                        break;

                    case "x":
                        curves.Push(new object[] { typeof(XCurve), body } );
                        break;

                    case "z":
                        curves.Push(new CartesianCurve(body));
                        break;

                    case "s":
                        curves.Push(new SetCurve(body));
                        break;

                    default:
                        throw new UsageException(Usage);
                }
            }
        }

        // fixup any lazily constructed curves if they didn't turn out to
        // be parametric after all
        List<Curve> list = new List<Curve>();

        foreach (object obj in curves) {
            Curve curve = obj as Curve;

            if (curve == null) {
                object[] data = (object[])obj;
                object[] args = { data[1] };
                Type type = (Type)data[0];

                curve = (Curve)type.GetConstructor(Type.GetTypeArray(args)).Invoke(args);
            }

            list.Add(curve);
        }

        list.Reverse();

        return list;
    }



    // figure out whether an expression is actually a range or parametric definition
    static bool ParseRange(ParseTree parseTree, string name, Pair<Pair<Function>> ranges, Stack curves)
    {
        ParseTreeOperator op = parseTree as ParseTreeOperator;

        if ((op == null) || (op.Operator != OpTable.Colon) || (op.Args.Length != 2))
            return false;

        Pair<Function> range = new Pair<Function>(new Function(op.Args[0], null),
                                                  new Function(op.Args[1], null));

        switch (name) {

            case null:
            case "x":
                // x axis range
                ranges[0] = range;
                break;

            case "y":
                // y axis range
                ranges[1] = range;
                break;

            case "t":
                // t range combines two previous expressions to make a parametric curve
                if (curves.Count < 2)
                    throw new UsageException(Usage);

                object[] y = curves.Pop() as object[];
                object[] x = curves.Pop() as object[];

                if ((x == null) || (x[0] != typeof(XCurve)) ||
                    (y == null) || (y[0] != typeof(YCurve)))
                    throw new UsageException(Usage);

                ParseTree xt = (ParseTree)x[1];
                ParseTree yt = (ParseTree)y[1];

                curves.Push(new ParametricCurve(xt, yt, range));
                break;

            default:
                throw new UsageException(Usage);
        }

        return true;
    }
}
